# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import, division, print_function, unicode_literals
import contextlib
import functools
import hashlib
import io
import itertools
import json
import os.path
import shutil
import tempfile
import time

import jinja2
import pip.baseparser
import pkg_resources
import yaml

from . import __version__

GROUP_SEPARATOR = b'\x1D'
RECORD_SEPARATOR = b'\x1E'
UNIT_SEPARATOR = b'\x1F'


def copy_resource(resource, destination, package='grocker'):
    resource_path = pkg_resources.resource_filename(package, resource)
    if pkg_resources.isdir(resource_path):
        shutil.copytree(resource_path, destination)
    else:
        shutil.copy2(resource_path, destination)


def load_yaml(file_path):
    with io.open(file_path, encoding='utf-8') as fp:
        return yaml.load(fp.read())


def load_yaml_resource(resource, package='grocker'):
    resource_path = pkg_resources.resource_filename(package, resource)
    return load_yaml(resource_path)


def get_run_dependencies(dependency_list):
    """
    Parse list of dependencies to only get run dependencies.

    Dependency list is a list of string or dict, which match the following
    format:

     [
        'run_dependency_1',
        {'run_dependency_2': 'build_dependency_2'},
        {'run_dependency_3': ['build_dependency_3.1', 'build_dependency_3.2']},
    ]
    """
    for dependency in dependency_list:
        if isinstance(dependency, dict):
            for key in dependency:
                yield key
        else:
            yield dependency


def get_build_dependencies(dependency_list):
    """
    Parse list of dependencies to only get build dependencies

    see get_run_dependencies() for dependency list format
    """
    for dependency in dependency_list:
        if isinstance(dependency, dict):
            for value_or_list in dependency.values():
                if isinstance(value_or_list, list):
                    for value in value_or_list:
                        yield value
                else:
                    yield value_or_list
        else:
            yield dependency


def get_dependencies(config, with_build_dependencies=False):
    runtime_dependencies = config['system']['runtime'][config['runtime']]

    dependencies = itertools.chain(
        config['system']['base'],
        get_run_dependencies(runtime_dependencies),
        get_run_dependencies(config['dependencies'])
    )

    if with_build_dependencies:
        build_dependencies = itertools.chain(
            config['system']['build'],
            get_build_dependencies(runtime_dependencies),
            get_build_dependencies(config['dependencies'])
        )

        dependencies = itertools.chain(dependencies, build_dependencies)

    return list(dependencies)


def config_identifier(config):
    """
    Hash config to get an unique identifier

    Args:
        config (dict): Grocker config

    Returns:
        str: Config identifier (SHA 256)
    """
    def unit_list(l):
        return UNIT_SEPARATOR.join(sorted(x.encode('utf-8') for x in l))

    dependencies = unit_list(get_dependencies(config, with_build_dependencies=True))
    repositories = RECORD_SEPARATOR.join(
        unit_list([name] + [cfg[x] for x in sorted(cfg)])
        for name, cfg in config['repositories'].items()
    )
    data = GROUP_SEPARATOR.join([
        dependencies,
        repositories,
    ])
    digest = hashlib.sha256(data)
    return digest.hexdigest()


def render_template(template_path, output_path, context):
    env = jinja2.Environment()
    env.filters['jsonify'] = json.dumps

    with io.open(template_path, encoding='utf-8') as template_file:
        template = env.from_string(template_file.read())

    output = template.render(**context)

    with io.open(output_path, mode='w', encoding='utf-8') as output_file:
        output_file.write(output)


def default_image_name(docker_image_prefix, release):
    req = pkg_resources.Requirement.parse(release)
    assert str(req.specifier).startswith('=='), "Only fixed version can use default image name."
    img_name = "{project}{extra_requirements}:{project_version}-{grocker_version}".format(
        project=req.project_name,
        extra_requirements='-' + '-'.join(req.extras) if req.extras else '',
        project_version=str(req.specifier)[2:],
        grocker_version=__version__,
    )
    return '/'.join((docker_image_prefix, img_name)) if docker_image_prefix else img_name


@contextlib.contextmanager
def pip_conf(pip_conf_path=None):
    """
    Use or fake (when it does not exist) a pip config file

    Args:
        pip_conf_path: str, pip configuration file to use if it exists

    Yield:
        str, the path to the pip configuration file (or the faked one)
    """
    if pip_conf_path is None or not os.path.exists(pip_conf_path):
        with tempfile.NamedTemporaryFile('w', dir=os.path.expanduser('~/.cache')) as f:
            config = pip.baseparser.ConfigOptionParser(name='global').config
            config.write(f)
            f.flush()
            yield f.name
    else:
        yield pip_conf_path


def retry(exception, tries=3, delay=1):
    def decorator(function):
        @functools.wraps(function)
        def inner(*args, **kwargs):
            for i in range(tries):
                try:
                    return function(*args, **kwargs)
                except exception:
                    if i + 1 == tries:
                        raise
                    time.sleep(delay)
        return inner
    return decorator


def get_version_from_requirement(requirement):
    if len(requirement.specifier) != 1:
        raise ValueError("Only exact specifier are accepted: %s" % requirement)
    spec = list(requirement.specifier)[0]
    if spec.operator not in ('==', '==='):
        raise ValueError("Only exact specifier are accepted: %s" % requirement)
    return spec.version



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import, division, print_function, unicode_literals
import pkg_resources

__version__ = pkg_resources.get_distribution('grocker').version
__copyright__ = '2015, Polyconseil'

DOCKER_API_VERSION = '1.21'



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

import contextlib
import functools
import itertools
import io
import json
import logging
import os.path
import re
import sys
import uuid

import docker
import docker.errors
import docker.utils
from packaging import requirements

from . import __version__, DOCKER_API_VERSION
from . import six
from . import helpers


DIGEST_RE = re.compile(r'[\w\.-]+: digest: (sha256:\w+) size: \d+')


def build_root_image(docker_client, config, tag=None):
    tag = tag or '{}.grocker'.format(uuid.uuid4())
    with six.TemporaryDirectory() as tmp_dir:
        build_dir = os.path.join(tmp_dir, 'build')
        helpers.copy_resource('resources/docker/root-image', build_dir)
        helpers.render_template(
            os.path.join(build_dir, 'Dockerfile.j2'),
            os.path.join(build_dir, 'Dockerfile'),
            {
                'image': config['system']['image'],
                'version': __version__,
                'runtime': config['runtime'],
            },
        )

        helpers.render_template(
            os.path.join(build_dir, 'apt-repositories.sh.j2'),
            os.path.join(build_dir, 'apt-repositories.sh'),
            {'repositories': config['repositories']},
        )

        dependencies = helpers.get_dependencies(config)
        build_env = {'SYSTEM_DEPS': ' '.join(dependencies)}
        return docker_build_image(docker_client, build_dir, tag=tag, buildargs=build_env)


def build_compiler_image(docker_client, root_image_tag, config, tag=None):
    tag = tag or '{}.grocker'.format(uuid.uuid4())
    with six.TemporaryDirectory() as tmp_dir:
        build_dir = os.path.join(tmp_dir, 'build')
        helpers.copy_resource('resources/docker/compiler-image', build_dir)
        helpers.render_template(
            os.path.join(build_dir, 'Dockerfile.j2'),
            os.path.join(build_dir, 'Dockerfile'),
            {
                'root_image_tag': root_image_tag,
                'runtime': config['runtime'],
            },
        )

        dependencies = helpers.get_dependencies(config, with_build_dependencies=True)
        with io.open(os.path.join(build_dir, 'provision.env'), 'w') as fp:
            fp.write('SYSTEM_DEPS="{}"'.format(' '.join(dependencies)))

        return docker_build_image(
            docker_client,
            build_dir,
            tag=tag,
            pull=bool(config['docker_image_prefix']),
        )


def build_pypi_image(docker_client, tag):
    with six.TemporaryDirectory() as tmp_dir:
        build_dir = os.path.join(tmp_dir, 'build')
        helpers.copy_resource('resources/docker/nginx-pypi', build_dir)
        return docker_build_image(docker_client, build_dir, tag=tag)


def get_or_create_data_volume(docker_client, name):
    volumes = docker_client.volumes()['Volumes'] or ()
    for volume in volumes:
        if volume['Name'] == name:
            return volume
    return docker_client.create_volume(name=name)


@contextlib.contextmanager
def http_wheel_server(docker_client, wheels_volume_name, config):
    nginx_image = docker_get_or_build_image(
        docker_client,
        config['docker_image_prefix'],
        'grocker-nginx-pypi:1.0.0',
        build_pypi_image,
    )
    nginx = docker_client.create_container(
        image=nginx_image,
        host_config=docker_client.create_host_config(
            binds={
                wheels_volume_name: {
                    'bind': '/wheels',
                    'mode': 'ro',
                }
            },
        ),
    )
    nginx_container_id = nginx.get('Id')
    docker_client.start(nginx_container_id)
    nginx_server_ip = docker_client.inspect_container(nginx_container_id)['NetworkSettings']['IPAddress']

    try:
        yield nginx_server_ip
    finally:
        remove_container = helpers.retry(docker.errors.APIError)(docker_client.remove_container)
        remove_container(nginx_container_id, force=True)


def build_runner_image(
        docker_client, root_image_tag, config, release, wheels_volume_name, tag=None
):
    tag = tag or '{}.grocker'.format(uuid.uuid4())
    requirement = requirements.Requirement(release)
    # Markers would not make much sense here and url are unsupported.
    assert requirement.marker is None, requirement
    assert requirement.url is None, requirement

    with http_wheel_server(docker_client, wheels_volume_name, config) as pypi_ip:
        with six.TemporaryDirectory() as tmp_dir:
            build_dir = os.path.join(tmp_dir, 'build')
            helpers.copy_resource('resources/docker/runner-image', build_dir)
            helpers.render_template(
                os.path.join(build_dir, 'Dockerfile.j2'),
                os.path.join(build_dir, 'Dockerfile'),
                {
                    'root_image_tag': root_image_tag,
                    'entrypoint_name': config['entrypoint_name'],
                    'app_name': requirement.name,
                    'app_extras': ','.join(sorted(requirement.extras)),
                    'app_version': helpers.get_version_from_requirement(requirement),
                    'volumes': config['volumes'],
                    'ports': config['ports'],
                },
            )

            # TODO(fbochu): .grocker file will be obsolete in next major version, so drop it.
            helpers.render_template(
                os.path.join(build_dir, '.grocker.j2'),
                os.path.join(build_dir, '.grocker'),
                {
                    'grocker_version': __version__,
                    'runtime': config['runtime'],
                    'release': release,
                },
            )

            if config['pip_constraint']:
                with io.open(config['pip_constraint'], 'r') as fp:
                    with io.open(os.path.join(build_dir, 'constraints.txt'), 'w') as f:
                        f.write(fp.read())

            return docker_build_image(
                docker_client,
                build_dir,
                tag=tag,
                pull=bool(config['docker_image_prefix']),
                buildargs={'GROCKER_PYPI_IP': pypi_ip},
            )


def get_root_image(docker_client, config):
    img_name = 'grocker-{runtime}-root:{version}-{hash}'.format(
        runtime=config['runtime'],
        version=__version__,
        hash=helpers.config_identifier(config),
    )
    return docker_get_or_build_image(
        docker_client,
        config['docker_image_prefix'],
        img_name,
        functools.partial(build_root_image, config=config),
    )


def get_compiler_image(docker_client, config):
    img_name = 'grocker-{runtime}-compiler:{version}-{hash}'.format(
        runtime=config['runtime'],
        version=__version__,
        hash=helpers.config_identifier(config),
    )
    root_tag = get_root_image(docker_client, config)
    return docker_get_or_build_image(
        docker_client,
        config['docker_image_prefix'],
        img_name,
        functools.partial(build_compiler_image, config=config, root_image_tag=root_tag),
    )


def get_pip_env(pip_conf):
    logger = logging.getLogger(__name__)

    def get(cfg, section, option, default=None):
        try:
            return cfg.get(section, option)
        except (six.configparser.NoSectionError, six.configparser.NoOptionError):
            return default

    if not pip_conf:
        return {}

    logger.info('-> Pip use configuration from %s.', pip_conf)
    config = six.configparser.ConfigParser()
    config.read(pip_conf)

    env = {
        'PIP_INDEX_URL': get(config, 'global', 'index-url'),
        'PIP_EXTRA_INDEX_URL': get(config, 'global', 'extra-index-url'),
    }
    env = {k: v for k, v in env.items() if v}
    logger.debug('pip using env: %s', env)
    return env


def compile_wheels(docker_client, compiler_tag, config, release, wheels_volume_name, pip_conf):
    wheels_destination_volume = get_or_create_data_volume(docker_client, wheels_volume_name)
    binds = {
        wheels_destination_volume['Name']: {
            'bind': '/home/grocker/packages',
            'mode': 'rw',
        },
    }

    command = ['--python', config['runtime'], release]

    if config['pip_constraint']:
        binds[os.path.abspath(config['pip_constraint'])] = {
            'bind': '/home/grocker/constraints.txt',
            'mode': 'ro',
        }
        command = ['--pip-constraint', '/home/grocker/constraints.txt'] + command

    docker_run_container(docker_client, compiler_tag, command, binds=binds, environment=get_pip_env(pip_conf))


def docker_get_client(**kwargs):
    # XXX: hack around docker-machine certificate issue https://github.com/docker/docker-py/issues/731
    extra_kwargs = {'assert_hostname': False} if 'DOCKER_MACHINE_NAME' in os.environ else {}
    all_extra_kwargs = docker.utils.kwargs_from_env(**extra_kwargs)
    all_extra_kwargs.update(kwargs)
    all_extra_kwargs['version'] = DOCKER_API_VERSION
    client = docker.Client(**all_extra_kwargs)
    client.version()  # Call the API, this will raise if API version is not compatible.
    return client


def docker_build_image(docker_client, path, tag=None, pull=True, buildargs=None):
    stream = docker_client.build(path=path, tag=tag, rm=True, forcerm=True, pull=pull, buildargs=buildargs)
    data = inspect_stream(stream)
    if not data['success']:
        raise RuntimeError('Image build failed')
    return tag


def inspect_stream(stream):
    """Return some data about the stream."""
    old_status = None
    data = {'success': True}
    for line in itertools.chain.from_iterable(six.smart_text(x).splitlines() for x in stream):
        line = json.loads(line)
        if 'status' in line:
            status = line['status']
            match = DIGEST_RE.match(status)
            if match:
                data['hash'] = match.group(1)
            if old_status != status:
                old_status = status
                print()
                print(old_status, end='')
            else:
                print('.', end='')
                sys.stdout.flush()  # Python 2.7 does not support flush argument on print() function.
        elif 'stream' in line:
            print(line['stream'], end='')
        elif 'error' in line:
            print(line['error'])
            data['success'] = False
        else:
            print(line)
    print()
    return data


def docker_pull_image(docker_client, name):
    logger = logging.getLogger(__name__)
    logger.info('Pulling %s...', name)

    stream = docker_client.pull(name, stream=True)
    inspect_stream(stream)
    return [image for image in docker_client.images() if name in (image['RepoTags'] or [])]


def docker_push_image(docker_client, name):
    logger = logging.getLogger(__name__)
    logger.info('Pushing %s...', name)

    stream = docker_client.push(name, stream=True)
    data = inspect_stream(stream)
    return data['hash']


def docker_get_or_build_image(docker_client, prefix, name, builder):
    full_name = '/'.join((prefix, name)) if prefix else name
    images = [image for image in docker_client.images() if full_name in (image['RepoTags'] or [])]
    if not images and prefix:
        images = docker_pull_image(docker_client, full_name)
    if not images:
        builder(docker_client, tag=full_name)
        if prefix:
            docker_push_image(docker_client, full_name)
    return full_name


def docker_run_container(docker_client, tag, command, binds=None, environment=None):
    container = docker_client.create_container(
        image=tag,
        command=command,
        environment=environment,
        volumes=[x['bind'] for x in binds.values()] if binds else [],
        host_config=docker_client.create_host_config(binds=binds),
    )

    container_id = container.get('Id')
    docker_client.start(container_id)
    stream = docker_client.attach(container_id, stream=True, logs=True)
    for line in stream:
        print(line.decode('utf-8'), end='')
    return_code = docker_client.wait(container_id)
    docker_client.remove_container(container_id)
    if return_code != 0:
        raise RuntimeError('Container exit with a non-zero return code (%d).', return_code)


def docker_purge_volumes(docker_client, filters=()):
    """Remove dangling volumes (volumes not referenced by any container).

    Args:
        docker_client: a docker-py Client.
        filters: see PurgeAction for choices.

    If ``build`` is in filters, all dangling volumes are deleted.
    If ``dangling`` is the *only* item of filters, the named volumes used by
    grocker as a cache are preserved.
    """
    if 'build' in filters or 'dangling' in filters:
        dangling_volumes = docker_client.volumes(filters={'dangling': True})['Volumes']
        if filters == ['dangling']:
            # Exclude wheels volumes that acts as a grocker cache
            volumes = []
            for volume in dangling_volumes:
                if 'grocker-wheels-cache' not in volume['Name']:
                    # Non grocker data volume => clean it.
                    volumes.append(volume)
                else:
                    # grocker data volume, clean it if outdated.
                    volume_grocker_version = volume['Name'].split('-')[3]
                    if __version__ < volume_grocker_version:
                        volumes.append(volume)
            dangling_volumes = volumes
        for volume in dangling_volumes:
            print('Purging volume: {name}'.format(name=volume['Name']))
            try:
                docker_client.remove_volume(volume['Name'])
            except docker.errors.APIError:
                logging.getLogger(__name__).error("Could not remove volume %s, skipping.", volume['Name'])


def docker_purge_images(docker_client, filters=()):
    filter_desc = {
        'dangling': {'dangling': True},
        'runners': {'label': 'grocker.image.kind=runner'},
        'builders': {'label': 'grocker.image.kind=builder'},
    }

    for filter_name in filters:
        images = docker_client.images(filters=filter_desc[filter_name])
        for image in images:
            print('Purging image: {tags} ({img[Id]})'.format(img=image, tags=', '.join(image['RepoTags'])))
            docker_client.remove_image(image)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import, division, print_function, unicode_literals
import argparse
import enum
# pylint: disable=wrong-import-order
import io
import logging
import os
import os.path
import sys
# pylint: enable=wrong-import-order

import yaml

from . import __version__
from . import builders
from . import helpers
from . import loggers


class GrockerActions(enum.Enum):
    build_dep = 'dep'  # -> dep
    build_img = 'img'  # -> img
    push_img = 'push'  # -> push
    all = 'build'  # -> dep + img + push


def arg_parser():
    """
    Create an CLI args parser

    Some default value (marked as #precedence) are set to None due to precedence
    order (see parse_config() doc string for more information).
    """
    def file_path_type(x):
        return os.path.abspath(os.path.expanduser(x))

    def file_path_or_none_type(x):
        return file_path_type(x) if x is not None else None

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c', '--config', action='append', type=file_path_or_none_type, default=[],
        help='Grocker config file',
    )
    parser.add_argument(  # precedence
        '-r', '--runtime', default=None,
        help="runtime used to build and run this image",
    )
    parser.add_argument(  # precedence
        '-e', '--entrypoint-name', default=None,
        help="Docker entrypoint to use to run this image",
    )
    parser.add_argument(  # precedence
        '--volume', action='append', default=[], dest='volumes',
        help="Container storage and configuration area",
    )
    parser.add_argument(  # precedence
        '--port', action='append', default=[], dest='ports',
        help="Port on which a container will listen for connections"
    )
    parser.add_argument(
        '--pip-conf', metavar='<file>', type=file_path_or_none_type, default=None,
        help="pip configuration file used to download dependencies (by default use pip config getter)",
    )
    parser.add_argument(
        '--pip-constraint', metavar='<file>', type=file_path_or_none_type, default=None,
        help="pip constraint file used to download dependencies",
    )
    parser.add_argument(
        '--docker-image-prefix', metavar='<url>',
        help='docker registry or account on Docker official registry to use'
    )
    parser.add_argument('-n', '--image-name', metavar='<name>', help="name used to tag the build image")
    parser.add_argument(
        'actions', choices=GrockerActions, type=GrockerActions, nargs='+',
        metavar='<action>', help='should be one of {}'.format(', '.join(x.value for x in GrockerActions))
    )

    parser.add_argument(
        '--result-file', metavar='<filename>', default=None,
        help="yaml file where results (image name, ...) are written"
    )
    parser.add_argument('release', metavar='<release>', help="application to build (you can use version specifier)")

    parser.add_argument('--purge', action=PurgeAction, help="purge docker images")
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose mode')

    return parser


class PurgeAction(argparse.Action):
    """Argparse action to purge grocker related docker images"""

    def __init__(self, option_strings, *args, **kwargs):
        super(PurgeAction, self).__init__(
            option_strings=option_strings,
            dest=argparse.SUPPRESS,
            choices=('all', 'builders', 'dangling', 'runners'),
        )

    def __call__(self, parser, namespace, values, *args, **kwargs):
        filters = [values] if values != 'all' else [x for x in self.choices if x != 'all']
        builders.docker_purge_images(builders.docker_get_client(), filters)
        builders.docker_purge_volumes(builders.docker_get_client(), filters)
        parser.exit()


def parse_config(config_paths, **kwargs):
    """
    Generate config regarding precedence order

    Precedence order is defined as :

    1. Command line arguments
    2. project ``.grocker.yml`` file (or the one specified on the command line)
    3. the grocker ``resources/grocker.yaml`` file
    """
    config = helpers.load_yaml_resource('resources/grocker.yaml')

    if not config_paths and os.path.exists('.grocker.yml'):
        config_paths = ['.grocker.yml']

    for config_path in config_paths:
        project_config = helpers.load_yaml(config_path)
        config.update(project_config)

    config.update({k: v for k, v in kwargs.items() if v})

    return config


def clean_actions(actions):
    if GrockerActions.all in actions:
        actions_without_all = set(GrockerActions) - {GrockerActions.all}
        if actions != [GrockerActions.all]:
            sys.exit(
                "The {all} action already specifies {actions_without_all}.\n"
                "Please use either {all} or any combination of ({actions_without_all}).".format(
                    all=GrockerActions.all.value,
                    actions_without_all=', '.join(sorted(action.value for action in actions_without_all)),
                )
            )
        actions = actions_without_all
    return actions


def main():
    parser = arg_parser()
    args = parser.parse_args()
    config = parse_config(
        args.config,
        runtime=args.runtime,
        entrypoint_name=args.entrypoint_name,
        pip_constraint=args.pip_constraint,
        docker_image_prefix=args.docker_image_prefix,
        volumes=args.volumes,
        ports=args.ports,
    )

    loggers.setup(verbose=args.verbose)
    logger = logging.getLogger('grocker' if __name__ == '__main__' else __name__)

    args.actions = clean_actions(args.actions)

    if config['runtime'] not in config['system']['runtime']:
        raise RuntimeError('Unknown runtime: %s', config['runtime'])

    docker_client = builders.docker_get_client()
    image_name = args.image_name or helpers.default_image_name(config['docker_image_prefix'], args.release)
    results = {
        'release': args.release,
        'image': image_name,
    }

    wheels_volume_name = 'grocker-wheels-cache-{version}-{hash}'.format(
        version=__version__,
        hash=helpers.config_identifier(config),
    )

    if GrockerActions.build_dep in args.actions:
        logger.info('Compiling dependencies...')
        compiler_tag = builders.get_compiler_image(docker_client, config)
        results['compiler_image'] = compiler_tag
        with helpers.pip_conf(pip_conf_path=args.pip_conf) as pip_conf:
            builders.compile_wheels(
                docker_client=docker_client,
                compiler_tag=compiler_tag,
                config=config,
                release=args.release,
                wheels_volume_name=wheels_volume_name,
                pip_conf=pip_conf,
            )

    if GrockerActions.build_img in args.actions:
        logger.info('Building image...')
        root_image_tag = builders.get_root_image(docker_client, config)
        results['root_image'] = root_image_tag
        builders.build_runner_image(
            docker_client=docker_client,
            root_image_tag=root_image_tag,
            config=config,
            release=args.release,
            wheels_volume_name=wheels_volume_name,
            tag=image_name,
        )

    if GrockerActions.push_img in args.actions:
        if '/' not in image_name:
            logger.warning('Not pushing any image since the registry is unclear in %s', image_name)
        else:
            logger.info('Pushing image...')
            results['hash'] = builders.docker_push_image(docker_client, image_name)

    if args.result_file:
        result_dir = os.path.dirname(args.result_file)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        with io.open(args.result_file, 'w') as fp:
            yaml.dump(results, fp, indent=True)

if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import, division, print_function, unicode_literals

import shutil
import tempfile
import types

# pylint: disable=unused-import,import-error
try:
    import configparser
except ImportError:
    import ConfigParser as configparser  # noqa
# pylint: enable=unused-import,import-error


def super6(cls, self, method, *args, **kwargs):
    """
    A kind of super() working on both old style and new style classes.
    """
    classobj = getattr(types, 'ClassType', type(None))
    if isinstance(cls, classobj):
        getattr(cls.__bases__[-1], method)(self, *args, **kwargs)
    else:
        getattr(super(cls, self), method)(*args, **kwargs)


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", directory=None):
        self.name = tempfile.mkdtemp(suffix, prefix, directory)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        shutil.rmtree(self.name)


def smart_text(text, encoding='utf-8'):
    return text.decode(encoding) if isinstance(text, bytes) else text



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.
from __future__ import absolute_import, print_function, unicode_literals
import argparse
import logging
import logging.config
import os.path
import subprocess
import tempfile

try:  # Python 3+
    import configparser
except ImportError:  # Python 2.7
    import ConfigParser as configparser


WHEELS_DIRECTORY = os.path.expanduser('~/packages')


def arg_parser():
    file_validator = os.path.expanduser

    parser = argparse.ArgumentParser()
    parser.add_argument('--python', default='python')
    parser.add_argument('--pip-constraint', type=file_validator)
    parser.add_argument('--no-color', action='store_true')
    parser.add_argument('release', nargs='+')

    return parser


def setup_logging(enable_colors):
    colors = {'begin': '\x1b[1;36m', 'end': '\x1b[0m'} if enable_colors else {'begin': '', 'end': ''}

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'simple': {'format': '{begin}%(message)s{end}'.format(**colors)},
        },
        'handlers': {
            'console': {'class': 'logging.StreamHandler', 'formatter': 'simple'},
        },
        'loggers': {
            __name__: {'handlers': ['console'], 'level': 'INFO'},
        },
    })


def info(*msg):
    logging.getLogger(__name__).info(*msg)


def setup_pip(venv, package_dir):
    """Generate venv pip.conf."""
    info('Setup pip...')

    # Standard config
    guest_config = configparser.ConfigParser()
    guest_config.add_section('global')
    guest_config.set('global', 'wheel-dir', package_dir)
    guest_config.set('global', 'find-links', package_dir)

    # Write config
    venv_pip_conf = os.path.join(venv, 'pip.conf')
    with open(venv_pip_conf, 'w') as f:
        guest_config.write(f)


def setup_venv(python):
    """Setup venv and return python interpreter."""
    info('Setup venv using %s...', python)
    venv = tempfile.mkdtemp(suffix='.venv')
    subprocess.check_call([python, '-m', 'virtualenv', '-p', python, venv])
    subprocess.check_call([os.path.join(venv, 'bin', 'pip'), 'install', '-U', 'pip', 'setuptools', 'wheel'])
    return venv


def build_wheels(venv, package, package_dir, constraint=None):
    info('Building wheels for %s...', package)
    pip = os.path.join(venv, 'bin', 'pip')
    constraint_args = ['--constraint', constraint] if constraint else []
    try:
        subprocess.check_call([pip, 'wheel', '--wheel-dir', package_dir] + constraint_args + [package])
        return True
    except subprocess.CalledProcessError as exc:
        info(str(exc))
        if exc.output:
            print(exc.output)
        return False


def main():
    parser = arg_parser()
    args = parser.parse_args()
    setup_logging(not args.no_color)

    venv = setup_venv(args.python)
    setup_pip(venv, WHEELS_DIRECTORY)

    # grocker is a valid sudoer for the command /usr/bin/chown
    return_code = subprocess.call(['sudo', '/bin/chown', 'grocker', WHEELS_DIRECTORY])
    if return_code != 0:
        logging.getLogger(__name__).error(
            "grocker could not become the owner of %s, check the sudoers file", WHEELS_DIRECTORY,
        )

    for release in args.release:
        if not build_wheels(venv, release, WHEELS_DIRECTORY, args.pip_constraint):
            exit(1)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.
from setuptools import setup, find_packages


def read(filename):
    with open(filename) as f:
        return f.read()


setup(
    name='grocker',
    version='4.3.2',
    description="Docker image builder",
    long_description=read('Readme.rst'),
    keywords='docker build packaging',
    url='http://github.com/polyconseil/grocker',
    author='Polyconseil',
    author_email='opensource+grocker@polyconseil.fr',
    packages=find_packages(where='src', exclude=('tests', 'docs')),
    package_dir={'': str('src')},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'docker-py!=1.10.*',
        'Jinja2',
        'setuptools>=18.0.1',
        'pip>=7.1.2',
        'pyyaml>=3.11',
        'packaging',
    ],
    extras_require={
        ":python_version == '2.7'": [
            'enum34',
        ],
    },
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Software Distribution',
    ],
    entry_points={
        'console_scripts': (
            'grocker = grocker.__main__:main',
        ),
    },
)
